package com.appiancorp.analytics.excel_export;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.DataFormat;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import com.appiancorp.suiteapi.common.Name;
import com.appiancorp.suiteapi.content.Content;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentFilter;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.knowledge.Document;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.knowledge.FolderDataType;
import com.appiancorp.suiteapi.process.exceptions.SmartServiceException;
import com.appiancorp.suiteapi.process.framework.AppianSmartService;
import com.appiancorp.suiteapi.process.framework.Input;
import com.appiancorp.suiteapi.process.framework.MessageContainer;
import com.appiancorp.suiteapi.process.framework.Required;
import com.appiancorp.suiteapi.process.framework.SmartServiceContext;
import com.appiancorp.suiteapi.process.palette.PaletteInfo;
import com.appiancorp.suiteapi.type.AppianType;
import com.appiancorp.suiteapi.type.DatatypeProperties;
import com.appiancorp.suiteapi.type.NamedTypedValue;
import com.appiancorp.suiteapi.type.TypeService;
import com.appiancorp.suiteapi.type.TypedValue;

/**
 * 
 * @author sathya.srinivasan
 * @date 06/Aug/2013
 * Smart service class to export a CDT array to Excel document.
 */
@PaletteInfo(paletteCategory = "Appian Smart Services", palette = "Document Generation") 
public class ExportCdtToExcel extends AppianSmartService {

	private static final Logger LOG = Logger.getLogger(ExportCdtToExcel.class);
	private final SmartServiceContext smartServiceCtx;

	private TypedValue[] cdt;
	private String[] cdt_fields_to_export;
	private Long excel_base_template;
	private String starting_cell;
	private Long sheet_number =0l;
	private String document_name_to_create;
	private Long document_save_directory;
	private Boolean include_header_row =false;
	private Long document_to_overwrite;
	private String[] cell_keys;
	private String[] cell_values;
	private Long output_document;
	private Boolean fail_on_error =false;
	//private Long excel_document;
	private String error;
	
    private SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("dd-MMM-yyyy");
    private SimpleDateFormat DATETIME_FORMAT = new SimpleDateFormat("dd-MMM-yyyy HH:mm:ss");
    private SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HH:mm:ss");
    
    private static String MISMATCH_INPUT_FIELDS="The input fields does not match the fields in the CDT";
	
	private TypeService _ts;
	private ContentService cs;

	@Override
	public void run() throws SmartServiceException {
		 Locale currentLocale = smartServiceCtx.getUserLocale() != null ? smartServiceCtx.getUserLocale() : smartServiceCtx.getPrimaryLocale();

         try {
           Workbook wb = new XSSFWorkbook();
           wb.createSheet("Sheet1");
           if (excel_base_template != null) {
   			Document xls = (Document) cs.getVersion(excel_base_template, ContentConstants.VERSION_CURRENT); 
   			//String documentPath = xls.getInternalFilename();
   			String documentPath = cs.getInternalFilename(xls.getCurrentContentId());
             FileInputStream fis = new FileInputStream(documentPath);
             wb = WorkbookFactory.create(fis);
             fis.close();
           }

           Sheet s = wb.getSheetAt(this.sheet_number.intValue());

          
           //If any cell mapping is entered, put those values here

           if (cell_keys != null && cell_values != null
               && cell_keys.length == cell_values.length) {
             for (int i = 0; i < cell_keys.length; i++) {
               Cell a = ExportHelper.getCell(s, cell_keys[i]);
               a.setCellValue(cell_values[i]);
             }
           }

           //Set the date format      
           String dateformat = "d/m/yyyy";
           if (currentLocale.equals(Locale.US)) {
             dateformat = "m/d/yyyy";
           }
           DataFormat df = wb.createDataFormat();
           CellStyle cellStyleDateAndTime = wb.createCellStyle();
           cellStyleDateAndTime.setDataFormat(df.getFormat(dateformat + " h:mm"));
           CellStyle cellStyleDate = wb.createCellStyle();
           cellStyleDate.setDataFormat(df.getFormat(dateformat));
           
            //If starting cell not specified, use A1
            starting_cell = StringUtils.isEmpty(starting_cell) ? "A1" : starting_cell;
            Cell reportStart = ExportHelper.getCell(s, starting_cell);
            Cell current = reportStart;           
           
            LOG.info (" Written to Cell");
            
            int colCount;
            DatatypeProperties dtp=null;
            Long typeId = 0l;
            NamedTypedValue[] props = null; 
            int[] columnIndexToExport =null;
            
            for (TypedValue tv : cdt){
    			try{
    				Object[] values = (Object[])tv.getValue();
    				typeId = tv.getTypeRef().getId();
    				dtp = _ts.getDatatypeProperties(typeId);
    				LOG.debug("Data type properties : " + dtp);
    				props = _ts.getInstanceProperties(typeId);
    				LOG.debug("Named Type Values : " + props);
    				
    				for(int i=0; i<props.length; i++){
    					LOG.debug("Prop : "+props[i].getName() + " type :" + props[i].getInstanceType().toString() + " value : "+ values[i]);
    				}   				
    				
    			}catch(Exception e){
    				LOG.error(e.getMessage());
    				if (fail_on_error)
    	            	 throw createException(e, "error.export.general", e.getMessage());
    			}
    		}
            
            
            
            //Create a int[] variable holding the index of the fields to fetch from the CDT.
            if (cdt_fields_to_export ==null || cdt_fields_to_export.length==0){
            	colCount = props.length;       		
            	LOG.info ("Writing all columns. Total cols: " +  colCount);
            	//If CDT is blank (Technically this should'nt happen) then return error
            	if (colCount ==0){
            		LOG.error("Column Count =0. Cannot proceed forward");
            		error="Column Count = 0. Cannot proceed forward";
            		
            		return;
            	}            		
            	columnIndexToExport = new int[colCount];
            	for (int i=0; i<props.length; i++){
            		columnIndexToExport[i]=i;
            	}
            	LOG.debug("Added all columns to index");
            }
            else {
            	colCount = cdt_fields_to_export.length;
            	columnIndexToExport = new int[colCount];
            	LOG.info ("Writing requested columns. Total columns to write : " + colCount);
            	for (int i=0; i<cdt_fields_to_export.length; i++){
            		int currentCol = findFieldIndex(props, cdt_fields_to_export[i]);
            		//If the user selects fields but sends blank, then return error.
            		if (currentCol <0){
            			LOG.error(MISMATCH_INPUT_FIELDS);
            			error = MISMATCH_INPUT_FIELDS;
            			if (fail_on_error)
       	            	  throw createException(new Throwable(MISMATCH_INPUT_FIELDS), "error.export.general", MISMATCH_INPUT_FIELDS);
           				return;
            		}
            		else{
            			columnIndexToExport[i] = currentCol;
            			LOG.debug("Added CDT index " + currentCol + " to Array sequence index " + i);
            		}
            	}
            }
            
            /*
             * If header row is to be included, then run the following
             */
            LOG.debug("Header row request : " + include_header_row);
            if (include_header_row !=null && include_header_row){
                 for (int i=0; i< colCount; i++){
                	 //Get the column name....
                	 LOG.debug ("Printing Column Headers");
                      current.setCellValue(props[columnIndexToExport[i]].getName() );
                      LOG.debug ("Column Header " + i);
                      Cell next = current.getRow().getCell(current.getColumnIndex() + 1);
                      if (next ==null) {
                           next = current.getRow().createCell(current.getColumnIndex() +1 );
                      }
                      current = next;
                 }
                
                 //Move to the next row to write detail values
                 current = ExportHelper.getCell(s, reportStart.getRowIndex() +1, reportStart.getColumnIndex());
               LOG.info (" Header row written"); 
            }
         
         
          /* Now write the rows into DB */
         
            //Keep track of the row
          int rowIndex =reportStart.getRowIndex() +1;
          if (include_header_row !=null && include_header_row)
        	  rowIndex++;
         
          for (TypedValue thisCdt : cdt) {
                    for (int i = 0; i < colCount; i++) {
  
                        // int indexOfFieldInCdt = columnIndexToExport[i];
                        
                         populateCellValue(thisCdt, props, current, columnIndexToExport[i], cellStyleDate );
                        
                         Cell next = current.getRow().getCell(current.getColumnIndex() + 1);
                         if (next ==null) {
                           next = current.getRow().createCell(current.getColumnIndex() +1 );
                        }
                         current = next;                                  
                    }
                   
                    //Go to next Row
                    current = ExportHelper.getCell(s, rowIndex++, reportStart.getColumnIndex());
              

               }
           
           
	        Document d = registerDocument();
	        LOG.info (" Document Created ");
	        File file = new File(d.getInternalFilename());
	        LOG.info (" File created");
	        FileOutputStream fos = new FileOutputStream(file);
	        wb.write(fos);
	        fos.close();
	        LOG.info (" File stream closed");
           
            d.setSize((int) file.length());
            cs.updateFields(d, new Integer[]{ Document.COLUMN_SIZE}, ContentConstants.UNIQUE_NONE);
            output_document = d.getId();
            LOG.info (" Document updated");
        
         } catch (Exception e) {
             LOG.error(e, e);
             if (fail_on_error)
            	 throw createException(e, "error.export.general", e.getMessage());
         }


	}
	
	/*
	 * Find index of the attribute in the CDT
	 */
	private int findFieldIndex(NamedTypedValue[] arrayList, String stringToFind){
		int returnValue = -1;
		for (int i=0; i<arrayList.length; i++){
			LOG.debug("Currently searching for string " + stringToFind + " - against property field " + arrayList[i]);
			if(arrayList[i].getName().equals(stringToFind)){
				returnValue = i;
				break;
			}
		}
		return returnValue;
	}

	public ExportCdtToExcel(SmartServiceContext smartServiceCtx, TypeService ts, ContentService cs) {
		super();
		this.smartServiceCtx = smartServiceCtx;
		this._ts = ts;
		this.cs = cs;
	}
	
	
	 private void populateCellValue(TypedValue rs, NamedTypedValue[] props,  Cell current, int column, CellStyle cellStyleDate ) throws Exception {
         if (props[column]==null){
        	 LOG.debug("Found Empty Property at Index " +column +". Skipping..");
        	 return;
         } 
         Date d = null;
         
         Object[] values = (Object[])rs.getValue();
         String s1 ="";
         try{
        	 s1 = values[column].toString();
         }catch(Exception ex){
        	 LOG.debug("Found Empty Property at Index " +column +" for Column " +props[column]+". Skipping..");
        	 return;
         }
        		 
         Long coltype = props[column].getInstanceType();
         switch (coltype.intValue()) {
			case AppianType.BOOLEAN:
	             current.setCellType(Cell.CELL_TYPE_STRING);
	             if (s1==null)
	            	 current.setCellValue("");
	             else	            	          
	            	 current.setCellValue(s1.equals("1") ? "Yes" : "No");
	             LOG.debug("Boolean Value written : " + s1);
	             break;
		 case AppianType.DATE:
			 d = (Date) values[column];
				if (d == null) {
	                return;
	           } else {
	                current.setCellStyle(cellStyleDate);
	                current.setCellValue( DATE_FORMAT.format(d));
	                LOG.debug("Value written : " + DATE_FORMAT.format(d));
	                return;
	           }
		 case AppianType.TIMESTAMP:
			 d = (Date) values[column];
				if (d == null) {
	                return;
	           } else {
	                current.setCellStyle(cellStyleDate);
	                current.setCellValue( DATETIME_FORMAT.format(d));
	                LOG.debug("Value written : " + DATETIME_FORMAT.format(d));
	                return;
	           }
		 case AppianType.TIME:
			d = (Date) values[column];
			if (d == null) {
                return;
           } else {
                current.setCellStyle(cellStyleDate);
                current.setCellValue( TIME_FORMAT.format(d));
                LOG.debug("Value written : " + TIME_FORMAT.format(d));
                return;
           }
	      case AppianType.DOUBLE:
	    	  if (values[column]==null) {
	                return;
	           } else {
	        	   Double doub = (Double)values[column];
	                current.setCellType(Cell.CELL_TYPE_NUMERIC);
	                current.setCellValue(doub);
	                LOG.debug("Double Value written : " + doub);
	                return;
	           }
	      case AppianType.INTEGER:
	           if (values[column]==null) {
	                return;
	           } else {
	        	   Long doub = (Long)values[column];
	                current.setCellType(Cell.CELL_TYPE_NUMERIC);
	                current.setCellValue(doub);
	                LOG.debug("Integer Value written : " + doub);
	                return;
	           }

      case AppianType.STRING:
           if (s1 == null) {
                return;
           } else if (StringUtils.isEmpty(s1)) {
                return;
           } else {
                current.setCellType(Cell.CELL_TYPE_STRING);
                current.setCellValue(s1);
                LOG.debug("String Value written : " + s1 );
                return;
           }
     
      default:
           LOG.warn("Error no handler for " + coltype);
           current.setCellType(Cell.CELL_TYPE_STRING);
           if (coltype.intValue() > AppianType.INITIAL_CUSTOM_TYPE)
           	LOG.debug("Cannot write nested CDT to Excel");
           else{
	           current.setCellValue(s1);
	           LOG.debug("Char Value written : " + s1 );
           }
           return;

		}

     }
	
	
	/*
     * Register Document
     */
    
       private Document registerDocument() throws Exception {

              String name = document_name_to_create;
              String extension = "xlsx";

              ContentFilter cf = new ContentFilter(ContentConstants.TYPE_DOCUMENT);
              cf.setName(name);
              cf.setExtension(new String[] { extension });

              Long docId = null;
              if (document_to_overwrite != null) {
                Document d = (Document) cs.getVersion(document_to_overwrite,
                    ContentConstants.VERSION_CURRENT);
                cs.createVersion(d, ContentConstants.UNIQUE_FOR_PARENT);
                docId = d.getId();
              } else {
                Content[] children = cs.getChildren(document_save_directory, cf,
                    ContentConstants.GC_MOD_NORMAL);
                if (children == null || children.length == 0) {
                  Document d = new Document();
                  d.setName(name);
                  d.setExtension(extension);
                  d.setSize(1);
                  d.setParent(document_save_directory);
                  d.setState(ContentConstants.STATE_ACTIVE_PUBLISHED);
                  docId = cs.create(d, ContentConstants.UNIQUE_FOR_PARENT);
                } else {
                  Document d = (Document) children[0];
                  cs.createVersion(d, ContentConstants.UNIQUE_FOR_PARENT);
                  docId = d.getId();
                }
              }

              if (docId != null) {
                Document d = cs.download(docId, ContentConstants.VERSION_CURRENT, false)[0];
                return d;
              }

              return null;
            }
    
       
       private SmartServiceException createException(Throwable t, String key, Object... args) {
              return new SmartServiceException.Builder(getClass(), t).userMessage(key,
                  args).build();
            }
      
      
      

	public void onSave(MessageContainer messages) {
	}

	public void validate(MessageContainer messages) {
	}

	@Input(required = Required.OPTIONAL)
	@Name("cdt")
	public void setCdt(TypedValue[] val) {
		this.cdt = val;
	}

	@Input(required = Required.ALWAYS)
	@Name("excel_base_template")
	@DocumentDataType
	public void setExcel_base_template(Long val) {
		this.excel_base_template = val;
	}

	@Input(required = Required.OPTIONAL)
	@Name("starting_cell")
	public void setStarting_cell(String val) {
		this.starting_cell = val;
	}

	@Input(required = Required.ALWAYS)
	@Name("sheet_number")
	public void setSheet_number(Long val) {
		this.sheet_number = val;
	}

	@Input(required = Required.OPTIONAL)
	@Name("document_name_to_create")
	public void setDocument_name_to_create(String val) {
		this.document_name_to_create = val;
	}

	@Input(required = Required.ALWAYS)
	@Name("document_save_directory")
	@FolderDataType
	public void setDocument_save_directory(Long val) {
		this.document_save_directory = val;
	}

	@Input(required = Required.OPTIONAL)
	@Name("include_header_row")
	public void setInclude_header_row(Boolean val) {
		this.include_header_row = val;
	}
	
	@Input(required = Required.OPTIONAL)
	@Name("fail_on_error")
	public void setFail_on_error(Boolean val) {
		this.fail_on_error = val;
	}

	@Input(required = Required.OPTIONAL)
	@Name("document_to_overwrite")
	@DocumentDataType
	public void setDocument_to_overwrite(Long val) {
		this.document_to_overwrite = val;
	}

	@Input(required = Required.OPTIONAL)
	@Name("cell_keys")
	public void setCell_keys(String[] val) {
		this.cell_keys = val;
	}

	@Input(required = Required.OPTIONAL)
	@Name("cell_values")
	public void setCell_values(String[] val) {
		this.cell_values = val;
	}
	
	@Input(required = Required.OPTIONAL)
	@Name("cdt_fields_to_export")
	public void setCdt_fields_to_export(String[] val) {
		this.cdt_fields_to_export = val;
	}
	
	 @Name("output_document")
     @DocumentDataType
     public Long getOutput_document() {
          return output_document;
     }	
/*
	@Name("excel_document")
	@DocumentDataType
	public Long getExcel_document() {
		return excel_document;
	}
*/
	@Name("error")
	public String getError() {
		return error;
	}

}
