package com.appiancorp.servlet.custom;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import org.apache.log4j.Logger;

import com.appiancorp.services.ServiceContext;
import com.appiancorp.suiteapi.common.LocalObject;
import com.appiancorp.suiteapi.common.ObjectTypeMapping;
import com.appiancorp.suiteapi.common.ServiceLocator;
import com.appiancorp.suiteapi.common.exceptions.InvalidGroupException;
import com.appiancorp.suiteapi.common.exceptions.InvalidVersionException;
import com.appiancorp.suiteapi.common.exceptions.PrivilegeException;
import com.appiancorp.process.expression.ExpressionEvaluationException;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.content.exceptions.InvalidContentException;
import com.appiancorp.suiteapi.personalization.GroupService;
import com.appiancorp.suiteapi.personalization.User;
import com.appiancorp.suiteapi.personalization.UserService;
import com.appiancorp.suiteapi.process.ProcessDesignService;

public class ExcelFormatter {

  private static final Logger LOG = Logger.getLogger(ExcelFormatter.class);
  private static final char ARRAY_DELIMITER = ';';
  private static final DateFormat INCOMING_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
  private static final DateFormat INCOMING_DATE_TIME_FORMAT = new SimpleDateFormat(
    "yyyy-MM-dd HH:mm:ss");
  private static final int ENGLISH = 1;
  private static final int SPANISH = 2;
  private static final int FRENCH = 3;
  private static final String PERCENTAGE_TOKEN = "percentage";
  private static final Map<String, Integer> languages = new HashMap<String, Integer>() {
    private static final long serialVersionUID = 1L;
    {
      put("en", 1);
      put("es", 2);
      put("fr", 3);
    }
  };

  public static String retrieveGroupNames(Long[] groupIds, ServiceContext sc)
    throws InvalidGroupException, PrivilegeException {
    GroupService gs = ServiceLocator.getGroupService(sc);
    StringBuilder groupNames = new StringBuilder();
    for (Long groupId : groupIds) {
      if (groupId != null) {
        groupNames.append(gs.getGroup(groupId).getGroupName());
        groupNames.append(ARRAY_DELIMITER);
      } else {
        groupNames.append(ARRAY_DELIMITER);
      }
    }
    if (groupNames.length() > 0) {
      groupNames.deleteCharAt(groupNames.length() - 1);
    }
    return groupNames.toString();
  }

  public static String retrieveUserNamesFormatted(String[] userNames, ServiceContext sc) {
    UserService us = ServiceLocator.getUserService(sc);
    StringBuilder userNamesFormatted = new StringBuilder();
    User u = null;
    for (String userName : userNames) {

      if (LOG.isDebugEnabled()) {
        LOG.debug("Username found: " + userName);
      }

      if (userName != null && userName != "" && userName.length() > 0 &&
        userName.trim().length() > 0) {
        u = us.getUser(userName);
        userNamesFormatted.append(u.getFirstName() + " " + u.getLastName());
        userNamesFormatted.append(ARRAY_DELIMITER);
      } else {
        userNamesFormatted.append(ARRAY_DELIMITER);
      }
    }
    if (userNamesFormatted.length() > 0) {
      userNamesFormatted.deleteCharAt(userNamesFormatted.length() - 1);
    }
    return userNamesFormatted.toString();
  }

  public static String formatDates(String[] incomingDates, int dateFormat, Locale locale)
    throws ParseException {

    if (incomingDates.length > 0) {
      //StringBuilder datesFormateed = new StringBuilder("\"");
      StringBuilder datesFormatted = new StringBuilder("");
      for (String incomingDate : incomingDates) {
        if (incomingDate != null && incomingDate != "") {
          datesFormatted.append(DateFormat.getDateInstance(dateFormat, locale).format(
            INCOMING_DATE_FORMAT.parse(incomingDate)));
          datesFormatted.append(ARRAY_DELIMITER);
        } else {
          datesFormatted.append(ARRAY_DELIMITER);
        }
      }
      if (datesFormatted.length() > 0) {
        datesFormatted.deleteCharAt(datesFormatted.length() - 1);
        //datesFormatted.append("\"");
      }
      return datesFormatted.toString();
    }
    return new String("");
  }

  public static String formatDateTimes(String[] incomingDateTimes, int dateFormat,
    int dateTimeFormat, Locale locale, TimeZone timeZone) throws ParseException {

    if (incomingDateTimes.length > 0) {
      DateFormat outputDateTimeFormat = DateFormat.getDateTimeInstance(dateFormat, dateTimeFormat,
        locale);
      outputDateTimeFormat.setTimeZone(timeZone);
      //StringBuilder dateTimesFormatted = new StringBuilder("\"");
      StringBuilder dateTimesFormatted = new StringBuilder("");
      for (String incomingDateTime : incomingDateTimes) {
        if (incomingDateTime != null && incomingDateTime != "") {
          dateTimesFormatted.append(outputDateTimeFormat.format(INCOMING_DATE_TIME_FORMAT.parse(incomingDateTime)));
          dateTimesFormatted.append(ARRAY_DELIMITER);
        } else {
          dateTimesFormatted.append(ARRAY_DELIMITER);
        }
      }
      if (dateTimesFormatted.length() > 0) {
        dateTimesFormatted.deleteCharAt(dateTimesFormatted.length() - 1);
        //dateTimesFormatted.append("\"");
      }
      return dateTimesFormatted.toString();
    }
    return new String("");
  }

  public static String retrieveLocalObjectFormatted(LocalObject[] localObjects, ServiceContext sc)
    throws InvalidGroupException, PrivilegeException, NumberFormatException,
    InvalidContentException, InvalidVersionException {

    if (LOG.isDebugEnabled()) {
      LOG.debug("Formatting Local Object");
    }

    Integer type = null;
    StringBuilder localObjectFormatted = new StringBuilder();
    for (LocalObject localObject : localObjects) {
      if (localObject != null) {
        type = localObject.getType();
        switch (type) {
        case ObjectTypeMapping.LO_USER:
          localObjectFormatted.append(retrieveUserNamesFormatted(
            new String[] {localObject.getStringId()}, sc));
          localObjectFormatted.append(ARRAY_DELIMITER);
          break;
        case ObjectTypeMapping.LO_GROUP:
          localObjectFormatted.append(retrieveGroupNames(
            new Long[] {new Long(localObject.getId())}, sc));
          localObjectFormatted.append(ARRAY_DELIMITER);
          break;
        case ObjectTypeMapping.LO_DOCUMENT:
        case ObjectTypeMapping.LO_FOLDER:
          localObjectFormatted.append(formatContentObject(
            new Long[] {new Long(localObject.getId())}, sc));
          localObjectFormatted.append(ARRAY_DELIMITER);
          break;
        default:
          break;
        }
      } else {
        localObjectFormatted.append(ARRAY_DELIMITER);
      }

    }
    if (localObjectFormatted.length() > 0) {
      localObjectFormatted.deleteCharAt(localObjectFormatted.length() - 1);
    }

    return localObjectFormatted.toString();
  }

  public static String formatDuration(String duration, ServiceContext sc, Locale locale) 
		  throws ExpressionEvaluationException {
	  
	ProcessDesignService pds = ServiceLocator.getProcessDesignService(sc);

	duration = pds.evaluateExpression("=tostring(tointervalds(" + duration + "))").getValue().toString();
    String formattedDuration = "";
    String[] durationItems = duration.split(":");

    Long days = durationItems[0].equals("") ? 0 : Long.parseLong(durationItems[0]);
    Long hours = durationItems[2].equals("") ? 0 : Long.parseLong(durationItems[2]);
    Long minutes = durationItems[3].equals("") ? 0 : Long.parseLong(durationItems[3]);
    Long seconds = durationItems[4].split("\\.")[0].equals("") ? 0
      : Long.parseLong(durationItems[4].split("\\.")[0]);
    Long miliseconds = durationItems[4].split("\\.")[1].equals("") ? 0
      : Long.parseLong(durationItems[4].split("\\.")[1]);

    Integer language = languages.get(locale.getLanguage());

    switch (language) {
    case ENGLISH:
      formattedDuration += days > 0 ? days + " days " : "";
      formattedDuration += hours > 0 ? hours + " hrs " : "";
      formattedDuration += minutes > 0 ? minutes + " mins " + seconds + "." + miliseconds +
        " secs" : seconds + "." + miliseconds + " secs";
      break;
    case SPANISH:
      formattedDuration += days > 0 ? days + " das " : "";
      formattedDuration += hours > 0 ? hours + " hrs " : "";
      formattedDuration += minutes > 0 ? minutes + " mins " + seconds + "." + miliseconds +
        " segs" : seconds + "." + miliseconds + " segs";
      break;
    case FRENCH:
      formattedDuration += days > 0 ? days + " jours " : "";
      formattedDuration += hours > 0 ? hours + " h " : "";
      formattedDuration += minutes > 0 ? minutes + " min " + seconds + "." + miliseconds + " secs"
        : seconds + "." + miliseconds + " secs ";
      break;
    default:
      formattedDuration += days > 0 ? days + " days " : "";
      formattedDuration += hours > 0 ? hours + " hrs " : "";
      formattedDuration += minutes > 0 ? minutes + " mins " + seconds + "." + miliseconds +
        " secs" : seconds + "." + miliseconds + " secs";
      break;
    }
    return formattedDuration;
  }

  public static String formatContentObject(Long[] contentIds, ServiceContext sc)
    throws InvalidContentException, InvalidVersionException, PrivilegeException {

    ContentService cs = ServiceLocator.getContentService(sc);
    StringBuffer formattedContent = new StringBuffer();
    for (Long contentId : contentIds) {

      if (contentId != null) {
        formattedContent.append(cs.getVersion(contentId, ContentConstants.VERSION_CURRENT)
          .getDisplayName());
        formattedContent.append(ARRAY_DELIMITER);
      } else {
        formattedContent.append(ARRAY_DELIMITER);
      }
    }
    if (formattedContent.length() > 0) {
      formattedContent.deleteCharAt(formattedContent.length() - 1);
    }

    return formattedContent.toString();
  }

  public static String formatIntegers(Long[] integers) {
    StringBuilder formattedIntegers = new StringBuilder();
    for (Long integer : integers) {
      if (integer != null) {
        formattedIntegers.append(integer);
        formattedIntegers.append(ARRAY_DELIMITER);
      } else {
        formattedIntegers.append(ARRAY_DELIMITER);
      }

    }

    if (formattedIntegers.length() > 0) {
      formattedIntegers.deleteCharAt(formattedIntegers.length() - 1);
    }

    return formattedIntegers.toString();
  }

  public static String formatDecimals(Double[] doubles, String format) {
    StringBuilder formattedDecimals = new StringBuilder();

    if (LOG.isDebugEnabled()) {
      LOG.debug("Format " + format);
    }

    for (Double aDouble : doubles) {
      if (aDouble != null) {
        formattedDecimals.append(format.equals(PERCENTAGE_TOKEN) ? (aDouble * 100.0) + "%"
          : aDouble);
        formattedDecimals.append(ARRAY_DELIMITER);
      } else {
        formattedDecimals.append(ARRAY_DELIMITER);
      }
    }
    if (formattedDecimals.length() > 0) {
      formattedDecimals.deleteCharAt(formattedDecimals.length() - 1);
    }
    return formattedDecimals.toString();
  }

  public static String formatString(String[] strings) {
    StringBuilder formattedString = new StringBuilder();

    for (String string : strings) {
      if (string != null && string != "") {
        formattedString.append(string.replaceAll("\r\n", "").replaceAll("\"", "'"));
        formattedString.append(ARRAY_DELIMITER);
      } else {
        formattedString.append(ARRAY_DELIMITER);
      }
    }
    if (formattedString.length() > 0 &&
      formattedString.charAt(formattedString.length() - 1) == ARRAY_DELIMITER) {
      formattedString.deleteCharAt(formattedString.length() - 1);
    }
    return formattedString.toString();
  }

  public static String formatBoolean(Long[] booleans) {
    StringBuilder formattedString = new StringBuilder();
    for (Long eachBoolean : booleans) {
      if (eachBoolean != null) {
        formattedString.append(eachBoolean == 0 ? "False" : "True");
        formattedString.append(ARRAY_DELIMITER);
      } else {
        formattedString.append(ARRAY_DELIMITER);
      }

    }
    if (formattedString.length() > 0) {
      formattedString.deleteCharAt(formattedString.length() - 1);
    }
    return formattedString.toString();
  }
}
