package com.appiancorp.servlet.custom;

import java.io.OutputStream;
import java.text.DateFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.HSSFColor;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.DataFormat;
import org.apache.poi.ss.usermodel.Font;

import com.appiancorp.analytics.excel_export.ExportHelper;
import com.appiancorp.process.expression.ExpressionEvaluationException;
import com.appiancorp.services.ASLIdentity;
import com.appiancorp.services.ServiceContext;
import com.appiancorp.services.WebServiceContextFactory;
import com.appiancorp.suiteapi.common.LocalObject;
import com.appiancorp.suiteapi.common.ObjectTypeMapping;
import com.appiancorp.suiteapi.common.ServiceLocator;
import com.appiancorp.suiteapi.common.exceptions.InvalidGroupException;
import com.appiancorp.suiteapi.common.exceptions.InvalidVersionException;
import com.appiancorp.suiteapi.common.exceptions.PrivilegeException;
import com.appiancorp.suiteapi.content.exceptions.InvalidContentException;
import com.appiancorp.suiteapi.process.analytics2.Column;
import com.appiancorp.suiteapi.process.analytics2.ProcessAnalyticsService;
import com.appiancorp.suiteapi.process.analytics2.ProcessReport;
import com.appiancorp.suiteapi.process.analytics2.ReportData;
import com.appiancorp.suiteapi.process.analytics2.ReportResultPage;
import com.appiancorp.suiteapi.type.AppianType;
import com.appiancorp.suiteapi.type.TypedValue;

/**
 * @author elizabeth.epstein Servlet to export report data to Excel.
 */

public class ExportReportDataToExcelServletv2 extends HttpServlet {

	private static final Logger LOG = Logger
			.getLogger(ExportReportDataToExcelServletv2.class);

	private static final String PARAM_REPORT_ID = "reportId";
	private static final String PARAM_FILENAME = "filename";
	private static final String PARAM_CONTEXT = "context";
	private static final String PARAM_START_INDEX = "startIndex";
	private static final String PARAM_BATCH_SIZE = "batchSize";

	private final int DATE_FORMAT = DateFormat.MEDIUM;

	private ReportData reportData;
	private HSSFWorkbook wb;
	private org.apache.poi.ss.usermodel.Sheet sheet;
	private CellStyle styleHeader;
	private CellStyle styleDate;
	private CellStyle styleDateTime;

	private ServiceContext sc;

	@Override
	protected void doGet(HttpServletRequest q, HttpServletResponse r) {

		sc = WebServiceContextFactory.getServiceContext(q);
		Locale currentLocale = sc.getLocale();
		ProcessAnalyticsService pas = ServiceLocator
				.getProcessAnalyticsService2(sc);

		// Parse Parameters
		Long myReportID = Long.parseLong(q.getParameter(PARAM_REPORT_ID));
		Long startIndex = new Long(0);
		Long batchSize = new Long(-1);
		try {
			startIndex = Long.parseLong(q.getParameter(PARAM_START_INDEX));
			batchSize = Long.parseLong(q.getParameter(PARAM_BATCH_SIZE));
		} catch (Exception e) {
			LOG.debug("Error parsing startIndex or batchSize, using defaults");
		}
		String myFilename = q.getParameter(PARAM_FILENAME);
		if (StringUtils.isBlank(myFilename)) {
			myFilename = "Appian_Data_Export";
		}

		// Assume semicolon separated list of contexts.
		String myContextString = q.getParameter(PARAM_CONTEXT);

		try (OutputStream fileOut = r.getOutputStream()) {

			r.setContentType("application/vnd.ms-excel");
			r.setHeader("Content-Disposition", "attachment; filename="
					+ myFilename + ".xls");

			// Create Workbook and set date formatting
			wb = new HSSFWorkbook();
			sheet = wb.createSheet("ExportedData");
			initWorkbookFormatting(currentLocale);

			// Prepare ReportData
			ProcessReport report = pas.getProcessReport(myReportID);
			reportData = report.getData();
			setReportContext(reportData, myContextString,
					((ASLIdentity) sc.getIdentity()).getIdentity());
			reportData.setBatchSize(batchSize.intValue());
			reportData.setStartIndex(startIndex.intValue());

			ReportResultPage reportResultPage = pas
					.getReportPageWithTypedValues(reportData);

			// Write to Workbook Cells
			int numColumns = writeExcelWorkbook(reportResultPage);

			// Done writing data. Resize columns nicely.
			for (int i = 0; i < numColumns; i++) {
				sheet.autoSizeColumn(i);
			}
			// Write to Output Stream
			wb.write(fileOut);

		} catch (Exception e) {
			LOG.error("Unexpected error writing data to excel file", e);
		}

	}

	/*
	 * Sets date format and header column format.
	 */
	private void initWorkbookFormatting(Locale l_) {
		String dateformat = "dd/mm/yyyy";
		if (l_.equals(Locale.US)) {
			dateformat = "mm/dd/yyyy";
		}
		DataFormat df = wb.createDataFormat();
		styleDateTime = wb.createCellStyle();
		styleDateTime.setDataFormat(df.getFormat(dateformat + " hh:mm"));
		styleDate = wb.createCellStyle();
		styleDate.setDataFormat(df.getFormat(dateformat));

		styleHeader = wb.createCellStyle();
		Font font = wb.createFont();
		font.setBoldweight(Font.BOLDWEIGHT_BOLD);
		styleHeader.setFont(font);
		styleHeader
				.setFillForegroundColor(HSSFColor.LIGHT_CORNFLOWER_BLUE.index);
		styleHeader.setFillPattern(HSSFCellStyle.SOLID_FOREGROUND);
	}

	/**
	 * Deals with Process Model ID, Process ID, and User Context. If It's User
	 * Context-expecting report and no context is passed in, the default calling
	 * context will be used.
	 *
	 * @param rd_
	 * @param myContextString_
	 * @param scUsername_
	 */
	private void setReportContext(ReportData rd_, String myContextString_,
			String scUsername_) {

		int ctxType = rd_.getContextType();

		// Check for user task report, set custom user or Service Context user.
		if (ctxType == ReportData.CONTEXT_TYPE_ATTRIBUTED_TO_USER
				|| ctxType == ReportData.CONTEXT_TYPE_TASK_OWNER) {

			String username = (myContextString_ == null) ? scUsername_
					: myContextString_;
			if (LOG.isDebugEnabled()) {
				LOG.debug("Adding  User Context to Task Report: " + username);
			}
			LocalObject usercontext = new LocalObject();
			usercontext.setType(ObjectTypeMapping.TYPE_USER);
			usercontext.setStringId(username);
			LocalObject[] ucarray = new LocalObject[1];
			ucarray[0] = usercontext;
			rd_.setContext(ucarray);

		} else if (myContextString_ != null) { // Use integer Context Type
			String[] contextIDs = myContextString_.split(";");
			Long[] cidArray = new Long[contextIDs.length];
			for (int i = 0; i < contextIDs.length; i++) {
				cidArray[i] = Long.parseLong(contextIDs[i]);
			}
			int numIDs = cidArray.length;

			if (ctxType == ReportData.CONTEXT_TYPE_PROCESS) {
				if (LOG.isDebugEnabled()) {
					LOG.debug("Adding Custom Context to Process Report: "
							+ numIDs + " IDs");
				}

				LocalObject[] pcarray = new LocalObject[numIDs];
				for (int i = 0; i < numIDs; i++) {
					LocalObject intContext = new LocalObject();
					intContext.setType(ObjectTypeMapping.TYPE_BPM_PROCESS);
					intContext.setId(cidArray[i]);
					pcarray[i] = intContext;
				}
				rd_.setContext(pcarray);
			} else if (ctxType == ReportData.CONTEXT_TYPE_PROCESS_MODEL) {
				if (LOG.isDebugEnabled()) {
					LOG.debug("Adding Custom Context to Process Model Report: "
							+ numIDs + " IDs");
				}
				LocalObject[] pmcarray = new LocalObject[numIDs];
				for (int i = 0; i < numIDs; i++) {
					LocalObject intContext = new LocalObject();
					intContext
							.setType(ObjectTypeMapping.TYPE_BPM_PROCESS_MODEL);
					intContext.setId(cidArray[i]);
					pmcarray[i] = intContext;
				}
				rd_.setContext(pmcarray);
			}
		}

	}

	/**
	 * Write Cell using crazy methods written by SS that i don't understand.
	 * Return total column count.
	 */
	private int writeExcelWorkbook(ReportResultPage rrp_) {
		Column[] columns = reportData.getColumns();
		Map<String, Integer> columnTypes = rrp_.getColumnTypes();

		Cell reportStart = ExportHelper.getCell(sheet, "a1");
		Cell current = reportStart;
		int displayedColumnCount = 0;
		// Include Header Row
		for (Column c : columns) {
			if (c.getShow()) {
				displayedColumnCount++;
				current.setCellValue(c.getName());
				current.setCellStyle(styleHeader);
				Cell next = current.getRow().getCell(
						current.getColumnIndex() + 1);
				if (next == null) {
					next = current.getRow().createCell(
							current.getColumnIndex() + 1);
				}
				current = next;
			}
		}
		current = ExportHelper.getCell(sheet, reportStart.getRowIndex() + 1,
				reportStart.getColumnIndex());

		int rowi = current.getRowIndex();
		int celli_s = reportStart.getColumnIndex();
		for (int i = 0; i < rrp_.getResults().length; i++) {
			HashMap rowData = (HashMap) rrp_.getResults()[i];
			int celli = celli_s;
			for (Column c : columns) {
				if (c.getShow()) {
					current = ExportHelper.getCell(sheet, rowi, celli);

					TypedValue val = (TypedValue) rowData.get(c.getStringId());
					Integer type = (Integer) columnTypes.get(c.getStringId());

					if (val != null && val.toString() != ""
							&& val.getValue() != null && val.getValue() != ""
							&& val.getValue().toString() != ""
							&& val.getValue().toString().trim() != "") {
						try {
							
							switch (type.intValue()) {
							
							case AppianType.GROUP:
								current.setCellValue(ExcelFormatter.retrieveGroupNames(new Long[] { (Long) val.getValue() }, sc));
								break;
							case AppianType.LIST_OF_GROUP:
								current.setCellValue(ExcelFormatter.retrieveGroupNames((Long[]) val.getValue(), sc));
								break;
							case AppianType.USERNAME:
								current.setCellValue(ExcelFormatter.retrieveUserNamesFormatted(new String[] { (String) val.getValue() }, sc));
								break;
							case AppianType.LIST_OF_USERNAME:
								current.setCellValue(ExcelFormatter.retrieveUserNamesFormatted((String[]) val.getValue(), sc));
								break;
							case AppianType.DATE:
								current.setCellValue(ExcelFormatter.formatDates(new String[] { val.getValue().toString() }, DATE_FORMAT, sc.getLocale()));
								break;
							case AppianType.LIST_OF_DATE:
								ArrayList<String> dateList = new ArrayList<String>();
								for (Object date : (Object[]) val.getValue()) {
									if (date != null) {
										dateList.add(date.toString());
									} else {
										dateList.add("");
									}
								}
								current.setCellValue(ExcelFormatter.formatDates((String[]) dateList.toArray(new String[0]), DATE_FORMAT, sc.getLocale()));
								break;
							case AppianType.TIMESTAMP:
								current.setCellValue(ExcelFormatter
										.formatDateTimes(new String[] { val.getValue().toString() }, DATE_FORMAT, DATE_FORMAT, sc.getLocale(), sc.getTimeZone()));
								break;
							case AppianType.LIST_OF_TIMESTAMP:
								ArrayList<String> dateTimeList = new ArrayList<String>();
								for (Object dateTime : (Object[]) val
										.getValue()) {
									if (dateTime != null) {
										dateTimeList.add(dateTime.toString());
									} else {
										dateTimeList.add("");
									}
								}
								current.setCellValue(ExcelFormatter.formatDateTimes(
										(String[]) dateTimeList.toArray(new String[0]), DATE_FORMAT, DATE_FORMAT, sc.getLocale(), sc.getTimeZone()));
								break;
							case AppianType.INTERVAL_D_S:
								current.setCellValue(ExcelFormatter.formatDuration(val.getValue().toString(), sc, sc.getLocale()));
								break;
							case AppianType.DOCUMENT:
							case AppianType.FOLDER:
								current.setCellValue(ExcelFormatter.formatContentObject(new Long[] { (Long) val.getValue() }, sc));
								break;
							case AppianType.USER_OR_GROUP:
							case AppianType.DOCUMENT_OR_FOLDER:
								current.setCellValue(ExcelFormatter.retrieveLocalObjectFormatted(new LocalObject[] { (LocalObject) val.getValue() }, sc));
								break;
							case AppianType.LIST_OF_DOCUMENT:
							case AppianType.LIST_OF_FOLDER:
								current.setCellValue(ExcelFormatter.formatContentObject((Long[]) val.getValue(), sc));
								break;
							case AppianType.LIST_OF_USER_OR_GROUP:
							case AppianType.LIST_OF_DOCUMENT_OR_FOLDER:
								current.setCellValue(ExcelFormatter.retrieveLocalObjectFormatted((LocalObject[]) val.getValue(),sc));
								break;
							case AppianType.LIST_OF_INTEGER:
								current.setCellValue(ExcelFormatter.formatIntegers((Long[]) val.getValue()));
								break;
							case AppianType.DOUBLE:
								current.setCellValue(ExcelFormatter.formatDecimals(new Double[] { (Double) val.getValue() }, c.getFormatToken()));
								break;
							case AppianType.LIST_OF_DOUBLE:
								current.setCellValue(ExcelFormatter.formatDecimals((Double[]) val.getValue(),c.getFormatToken()));
								break;
							case AppianType.STRING:
								current.setCellValue(ExcelFormatter.formatString(new String[] { val.getValue().toString() }));
								break;
							case AppianType.LIST_OF_STRING:
								current.setCellValue(ExcelFormatter.formatString((String[]) val.getValue()));
								break;
							case AppianType.BOOLEAN:
								current.setCellValue(ExcelFormatter.formatBoolean(new Long[] { (Long) val.getValue() }));
								break;
							case AppianType.LIST_OF_BOOLEAN:
								current.setCellValue(ExcelFormatter.formatBoolean((Long[]) val.getValue()));
								break;
							default:
								if (LOG.isDebugEnabled()) {
									LOG.debug("Type didn't match the listed types: " + type.intValue()
											+ " Value is: " + val.getValue()
											+ " Format is: " + c.getFormatToken());
								}
								current.setCellValue(val.getValue().toString());
								break;
							}
						} catch (PrivilegeException | InvalidGroupException | InvalidContentException | 
								InvalidVersionException | ParseException | NumberFormatException | 
								ExpressionEvaluationException e) {
							LOG.error("Error in column " + c.getName() + " row " + i, e);
						}
					} else {
						current.setCellValue("");
					}

					Cell next = current.getRow().getCell(current.getColumnIndex() + 1);
					if (next == null) {
						next = current.getRow().createCell(current.getColumnIndex() + 1);
					}
					current = next;
					celli++;
				}
			}
			rowi++;

		} // End iteration
		return displayedColumnCount;
	}

}
